/* arbre.c (Combinatorial Pyramids project)
 *  Copyright (C) 2003 Luc Brun
 *  email: luc.brun@univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include  <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include "arbre.h"
#include "macros.h"
#include "coul_freq.h"

/*#define DEBUG 1*/
#define MAX(a,b) ((a)>(b)?(a):(b))
typedef double moments_fils[2][3];
/******************************************************
fonctions permettant de choisir
* comment couper sur l'axe 
* l'axe a couper
* le cluster a couper
*********************************************************/
typedef int     (*PFdecoupe)(arbre_binaire,coul_freq,double*,moments_fils,moments_fils);
typedef void    (*PFinit)(arbre_binaire,coul_freq,double*, double*,double*);
typedef int    (*PFaxe)(arbre_binaire,coul_freq);
typedef double (*PFcluster)(arbre_binaire);

#define Fils        SELF(fils)
#define Inter       SELF(inter)
#define Var         SELF(var)
#define Moy_axe     SELF(moy_axe)
#define Ecart       SELF(ecart)
#define Axe_decoupe SELF(axe_decoupe)
#define Moment0     SELF(moment0)
#define Moment1     SELF(moment1)
#define Moment2     SELF(moment2)
#define Numero      SELF(numero)
#define Feuille     SELF(feuille)
#define Choix_axe   SELF(choix_axe)
#define Valeur      SELF(valeur)
#define Partition   SELF(partition)

static  int private_decoupe(arbre_binaire,
                            coul_freq,
                            double*,
                            moments_fils,
                            moments_fils);
static  int private_decoupe_projette(arbre_binaire,
                                     coul_freq,
                                     double*,
                                     moments_fils,
                                     moments_fils);
static  int private_decoupe_median(arbre_binaire,
                                   coul_freq,
                                   double*,
                                   moments_fils,
                                   moments_fils);
static  int private_decoupe_median_dangereux(arbre_binaire,
                                             coul_freq,
                                             double*,
                                             moments_fils,
                                             moments_fils);

static void  private_initialisation(arbre_binaire,
                                    coul_freq,
                                    double*,
                                    double*,
                                    double*);
static void  private_init_dangereux(arbre_binaire,
                                    coul_freq,
                                    double*, 
                                    double*,
                                    double*);

static double private_eqm_pondere(arbre_binaire);
static double private_eqm_non_pondere(arbre_binaire);
static double private_cardinal(arbre_binaire);

static int private_choix_split_var(arbre_binaire,
                                   coul_freq);
static  int    private_choix_var(arbre_binaire,
                                 coul_freq);
static  int    private_choix_longueur(arbre_binaire,
                                      coul_freq);
static  int    private_choix_var_long(arbre_binaire,
                                      coul_freq);
static  int    private_choix_var_median(arbre_binaire,
                                        coul_freq);
static  int    private_choix_eqm(arbre_binaire,
                                 coul_freq);

static arbre_binaire  private_creer_fils(arbre_binaire);
static void           private_creer_table(arbre_binaire,
                                          coul_freq);

static void private_etiqueter(arbre_binaire,
                              int *,
                              int*);
static double private_decoupe_choix(arbre_binaire,
                                    coul_freq);
/* definition du type arbre */
struct arbre_binaire 
{
  arbre_binaire fils[2];
  intervalle inter[3];
  double var[3];
  int moy_axe[3];
  double ecart;
  int axe_decoupe;
  double moment0;
  double moment1[3];
  double moment2[3];
  int numero;
  bouleen feuille;
  PFaxe choix_axe;
  PFcluster valeur;
  PFdecoupe partition;
};

arbre_binaire FAIRE_arbre_binaire(coul_freq tab_freq,
                                  int cluster_choisi,
                                  int axe_choisi,
                                  int init_choisi,
                                  int decoupe_choisie)
{
  int i;
  double mom0,mom1[3],mom2[3];
  int binf[3],bsup[3];

  static PFaxe choix_de_l_axe[] = 
	{
		private_choix_var,
		private_choix_longueur,
		private_choix_var_long,
		private_choix_var_median,
		private_choix_eqm,
		private_choix_split_var	
	};
	static PFcluster valeur_cluster[] = 
	{
		private_eqm_pondere,
		private_eqm_non_pondere,
		private_cardinal
	};
	static PFinit choix_init[] = 
	{
		private_initialisation,
		private_init_dangereux
	};
  static PFdecoupe choix_decoupe[] = 
	{
		private_decoupe,
		private_decoupe_projette,
		private_decoupe_median,
		private_decoupe_median_dangereux
	};
  PFinit init = choix_init[init_choisi];

  INSTANCIER(arbre_binaire)

  Partition = choix_decoupe[decoupe_choisie];
  Valeur = valeur_cluster[cluster_choisi];
  Choix_axe = choix_de_l_axe[axe_choisi];



  binf[0] = coul_freq_give_min_C1(tab_freq);
  bsup[0] = coul_freq_give_max_C1(tab_freq)+1;
  binf[1] = coul_freq_give_min_C2(tab_freq);
  bsup[1] = coul_freq_give_max_C2(tab_freq)+1;
  binf[2] = coul_freq_give_min_C3(tab_freq);
  bsup[2] = coul_freq_give_max_C3(tab_freq)+1;


  Ecart =0.0;
  Fils[0] = (arbre_binaire)NULL;
  Fils[1]= (arbre_binaire)NULL;


  /* On fixe les intervalles */
	for(i=0;i<3;i++)
	{
		Inter[i].borne_inf = binf[i];
		Inter[i].borne_sup = bsup[i];
	}
#ifdef DEBUG
	printf("RACINE : variances :\n");
#endif
	/* Calcul des differents moments */
	init(self,tab_freq,&mom0,mom1,mom2);

	Moment0 = mom0;
	for(i=0;i<3;i++)
	{
		Moment1[i] = mom1[i];
		Moment2[i] = mom2[i];
		Moy_axe[i] = mom1[i]/mom0;
		Var[i]=(mom2[i]-mom1[i]*mom1[i]/mom0)/mom0;

#ifdef DEBUG
	printf("sur l'axe %d : %f\n",i,Var[i]);
#endif

	}
	Ecart =  Valeur(self);
	Axe_decoupe = Choix_axe(self,tab_freq);

#ifdef DEBUG
	printf("\n\n");
#endif
	Feuille = vrai;

	return(self);
}

void DEFAIRE_arbre_binaire(arbre_binaire self)
{
	if(Feuille == vrai)
	{
		free(self);
	}
	else
	{
		DEFAIRE_arbre_binaire(Fils[0]);
		DEFAIRE_arbre_binaire(Fils[1]);
	}
}

void  arbre_binaire_creer_la_table(arbre_binaire self,
				   coul_freq  tab_freq,
				   int* nb_de_couleur,
				   double ecart_max)

{
  int nb_couleur=1;
  double ecart1,ecart2,ecart_initial;
  bouleen fin;
  
  ecart_initial = Ecart;
  ecart1 = 100.0;
  fin = faux;
  
  do
    {
      
      private_creer_table(self,tab_freq);
      nb_couleur =nb_couleur + 1;
      ecart2 = (100*Ecart)/ecart_initial;
      /* on s' arrete si on est pres de ecart max et si l' */
      /* ecart intra classe decroit lentement*/
      fin = (ecart1-ecart2< 2.0)&&(ecart2 < ecart_max+2.0);
      /* On s' arrete de toute facon si on est en dessous de */
      /* ecart max*/
      fin = fin || (ecart2 <= ecart_max);
      ecart1 = ecart2;
#ifdef DEBUG
      printf("nb de couleurs = %d\n",nb_couleur);
      printf("ecart max dans l' arbre :%f\n",ecart2);
#endif
    }
  while(nb_couleur < *nb_de_couleur);
  /*    while(!fin);*/
  *nb_de_couleur = nb_couleur;
}
/*************************************************************************************/
/****** retourne l'indice du cluster d'une couleur donnee */
/*************************************************************************************/
void arbre_binaire_chercher_couleur_ds_table(arbre_binaire self,
                                             int* coul,
                                             int* indice_couleur)

{
	if(Feuille)
		*indice_couleur = Numero;
	else
	{
		if(coul[Axe_decoupe] < ((Fils[0]->inter[Axe_decoupe].borne_sup)*REDUCTION))
			arbre_binaire_chercher_couleur_ds_table(Fils[0],coul,indice_couleur);
		else
			arbre_binaire_chercher_couleur_ds_table(Fils[1],coul,indice_couleur);
	}
}

void arbre_binaire_etiqueter_feuilles(arbre_binaire self,int*  table)
{
  int numero = 0;
  
  
  private_etiqueter(self,table,&numero);
  
}

    /********************************************************************/
   /* Calcul de l'endroit ou il convient de couper sur l'axe choisi,   */
  /* ainsi que des moments associes aux deux clusters induits par le  */
 /* decoupage                                                        */
/********************************************************************/
static int private_decoupe(arbre_binaire self,
			   coul_freq tab_freq,
			   double* moment0,
			   moments_fils moment1,
			   moments_fils moment2)

{
  float S0 = Moment0;
  int axe1 = (Axe_decoupe+1)%3;
  int axe2 = (axe1 + 1)%3;
  int b_inf1 = Inter[Axe_decoupe].borne_inf;
  int b_sup1 = Inter[Axe_decoupe].borne_sup;

#ifdef DEBUG

  int b_inf2 = Inter[axe1].borne_inf;
  int b_sup2 = Inter[axe1].borne_sup;
  int b_inf3 = Inter[axe2].borne_inf;
  int b_sup3 = Inter[axe2].borne_sup;
  
#endif
  int i,j; /* variables de boucle */
  int indice_decoupe;
  double mom0,mom1[3],mom2[3];
  double  En[3],En2[3];
  double sn,terme,ui,freq_axei;
  double max = 0.0;
  double moments[3][2];
  double EQM;
  int demi_intervalle = (REDUCTION>>1);
  bouleen fin = faux;
  double X_moins_ya,det2;
  double a[3];
  double b2_par_delta2,delta2=0.0;
  /* initialisations */
  /************************************/
  /* sn : cardinal de Ct */
  /* En[], En2[] moments d'ordre 1 et 2*/
  /* a : moyenne du cluster */
  /* delta2 norme au carre du verteur tel que delta_i = longueur de */
  /* l'intervalle i */
  /************************************/
  sn=0.0;
  for(j=0;j<3;j++)
    {
      En[j]=0.0;
      En2[j] =0.0;
      a[j] = Moment1[j]/S0;
      delta2 += (Inter[j].borne_sup-Inter[j].borne_inf)*(Inter[j].borne_sup-Inter[j].borne_inf);
    }
  delta2 = delta2*REDUCTION*REDUCTION;
  b2_par_delta2 = S0*(S0/delta2);
  i= b_inf1;

#ifdef DEBUG
  printf("sur l'axe I%d : [%d,%d[\n",Axe_decoupe+1,b_inf1,b_sup1);
  printf("sur l'axe I%d : [%d,%d[\n",axe1+1,b_inf2,b_sup2);
  printf("sur l'axe I%d : [%d,%d[\n",axe2+1,b_inf3,b_sup3);
#endif

  /* Calculs */
  while((i<b_sup1)&&(!fin))
    {
      freq_axei= coul_freq_projection(tab_freq,Axe_decoupe,i,Inter,moments[axe1],moments[axe2]);
      
      
      ui = (i<<K_REDUCTION);
      if(i<0)
	ui -= demi_intervalle;
      else
	ui += demi_intervalle;
      
      sn += freq_axei;
      terme = ui*freq_axei;
      
      En[Axe_decoupe]   += terme;
      En[axe1] += moments[axe1][0];
      En[axe2] += moments[axe2][0];
      En2[Axe_decoupe] +=  ui*terme;
      En2[axe1] += moments[axe1][1];
      En2[axe2] += moments[axe2][1];
      EQM=0.0;
      if((sn != 0.0)&&(sn != S0))
	{
	  for(j=0;j<3;j++)
	    {
	      X_moins_ya = En[j]/sn-a[j];
	      EQM += X_moins_ya*X_moins_ya;
	    }
	  EQM = (EQM/(S0-sn))*sn;
	}
      
#ifdef DEBUG
      printf("EQM(%d) = %f, z= %f\n",i,EQM,sn/S0);
#endif
      
      if(max <  EQM)
	{
	  max = EQM;
	  indice_decoupe = i+1;
	  mom0 = sn;
	  det2 =b2_par_delta2*max;
	  for(j=0;j<3;j++)
	    {
	      mom1[j] = En[j];
	      mom2[j] = En2[j];
	    }
	}
      fin = ((sn+sn) > S0)&&(sn*(S0-sn) <= det2);
      i++;
    }
  /* transposition des resultats dans les variables de sorties */
  moment0[0]     = mom0;
  moment0[1]     = S0-mom0;
  for(j=0;j<3;j++)
    {
      moment1[0][j]     = mom1[j];
      moment1[1][j]     = Moment1[j]-mom1[j];
      moment2[0][j]     = mom2[j];
      moment2[1][j]     = Moment2[j]-mom2[j];
    }
#ifdef DEBUG
  printf("en sortie de boucle i = %d\n",i);
  printf("gain= %d\%\n",(int)(100*(b_sup1-i)/(b_sup1-b_inf1)));
#endif
  return(indice_decoupe);
}

static int private_decoupe_projette(arbre_binaire self,
				    coul_freq tab_freq,
				    double* moment0,
				    moments_fils moment1,
				    moments_fils moment2)

{
  double S0 = Moment0;
  int axe1 = (Axe_decoupe+1)%3;
  int axe2 = (axe1 + 1)%3;
  int b_inf1 = Inter[Axe_decoupe].borne_inf;
  int b_sup1 = Inter[Axe_decoupe].borne_sup;

#ifdef DEBUG

  int b_inf2 = Inter[axe1].borne_inf;
  int b_sup2 = Inter[axe1].borne_sup;
  int b_inf3 = Inter[axe2].borne_inf;
  int b_sup3 = Inter[axe2].borne_sup;

#endif

  int i,j; /* variables de boucle */
  int indice_decoupe;
  double mom0,mom1[3],mom2[3];
  double  En[3],En2[3];
  double sn,terme,ui,freq_axei;
  double max = 0.0;
  double moments[3][2];
  double EQM;
  int demi_intervalle = (REDUCTION>>1);
  bouleen fin = faux;
  double X_moins_ya,det2;
  double a;
  double b2_par_delta2,delta2=0.0;
  /* initialisations */
  /************************************/
  /* sn : cardinal de Ct */
  /* En[], En2[] moments d'ordre 1 et 2*/
  /* a : moyenne du cluster */
  /* delta2 norme au carre du verteur tel que delta_i = longueur de */
  /* l'intervalle i */
  /************************************/
  sn=0.0;
  for(j=0;j<3;j++)
    {
      En[j]=0.0;
      En2[j] =0.0;
      
    }
  a = Moment1[Axe_decoupe]/S0;
  delta2 = (Inter[Axe_decoupe].borne_sup-Inter[Axe_decoupe].borne_inf);
  delta2 = (delta2*REDUCTION)*(delta2*REDUCTION);
  b2_par_delta2 = S0*(S0/delta2);
  i= b_inf1;
  b_sup1 = (int)(b_sup1+a)/2; /* cf wong */
#ifdef DEBUG
  printf("sur l'axe I%d : [%d,%d[\n",Axe_decoupe+1,b_inf1,b_sup1);
  printf("sur l'axe I%d : [%d,%d[\n",axe1+1,b_inf2,b_sup2);
  printf("sur l'axe I%d : [%d,%d[\n",axe2+1,b_inf3,b_sup3);
#endif
  /* Calculs */
  while((i<b_sup1)&&(!fin))
    {
      freq_axei= coul_freq_projection(tab_freq,
				      Axe_decoupe,
				      i,
				      Inter,
				      moments[axe1],
				      moments[axe2]);
		
      
      ui = (i<<K_REDUCTION);
      if(i<0)
	ui -= demi_intervalle;
      else
	ui += demi_intervalle;

      sn += freq_axei;
      terme = ui*freq_axei;

      En[Axe_decoupe]   += terme;
      En[axe1] += moments[axe1][0];
      En[axe2] += moments[axe2][0];
      En2[Axe_decoupe] +=  ui*terme;
      En2[axe1] += moments[axe1][1];
      En2[axe2] += moments[axe2][1];

      EQM=0.0;
      if((sn != 0.0)&&(sn != S0))
	{
	  
	  X_moins_ya = En[Axe_decoupe]-sn*a;
	  EQM = X_moins_ya*X_moins_ya;
	  EQM = EQM/(sn*(S0-sn));
	}
      if(max <=  EQM)
	{
	  max = EQM;
	  indice_decoupe = i+1;
	  mom0 = sn;
	  det2 =b2_par_delta2*max;
	  for(j=0;j<3;j++)
	    {
	      mom1[j] = En[j];
	      mom2[j] = En2[j];
	    }
	}
      fin = (sn == S0);
/*(sn+sn) > S0)&&(sn*(S0-sn) <= det2);*/
      i++;
    }
  /* transposition des resultats dans les variables de sorties */
  moment0[0]     = mom0;
  moment0[1]     = S0-mom0;
  for(j=0;j<3;j++)
    {
      moment1[0][j]     = mom1[j];
      moment1[1][j]     = Moment1[j]-mom1[j];
      moment2[0][j]     = mom2[j];
      moment2[1][j]     = Moment2[j]-mom2[j];
    }
#ifdef DEBUG
  printf("en sortie de boucle i = %d\n",i);
  printf("gain= %d%\n",(int)(100*(b_sup1-i)/(b_sup1-b_inf1)));
#endif
  return(indice_decoupe);
}

static int private_decoupe_median(arbre_binaire self,
				  coul_freq tab_freq,
				  double* moment0,
				  moments_fils moment1,
				  moments_fils moment2)
{
  float S0 = Moment0;
  int axe1 = (Axe_decoupe+1)%3;
  int axe2 = (axe1 + 1)%3;
  int b_inf1 = Inter[Axe_decoupe].borne_inf;
  int b_sup1 = Inter[Axe_decoupe].borne_sup;

#ifdef DEBUG

  int b_inf2 = Inter[axe1].borne_inf;
  int b_sup2 = Inter[axe1].borne_sup;
  int b_inf3 = Inter[axe2].borne_inf;
  int b_sup3 = Inter[axe2].borne_sup;

#endif

  int i,j; /* variables de boucle */
  double moments[3][2];
  bouleen fin = faux;
  double  En[3],En2[3];
  double sn,terme,ui,freq_axei;
  int demi_intervalle = (REDUCTION>>1);

/************************************/  
  /* initialisations */
/************************************/
  
  i= b_inf1;
  sn =0.0;
    for(j=0;j<3;j++)
      {
	En[j]=0.0;
	En2[j] =0.0;
      }
#ifdef DEBUG

  printf("sur l'axe I%d : [%d,%d[\n",Axe_decoupe+1,b_inf1,b_sup1);
  printf("sur l'axe I%d : [%d,%d[\n",axe1+1,b_inf2,b_sup2);
  printf("sur l'axe I%d : [%d,%d[\n",axe2+1,b_inf3,b_sup3);

#endif
  /* Calculs */
  while((i<b_sup1-1)&&(!fin))
    {
      freq_axei= coul_freq_projection(tab_freq,Axe_decoupe,i,Inter,moments[axe1],moments[axe2]);
      

      ui = (i<<K_REDUCTION);
      if(i<0)
	ui -= demi_intervalle;
      else
			ui += demi_intervalle;
      
      sn += freq_axei;
      terme = ui*freq_axei;
      moments[Axe_decoupe][0] = terme;
      moments[Axe_decoupe][1] = ui*terme;
      for(j=0;j<3;j++)
	{
	  En[j]   += moments[j][0];
	  En2[j]  += moments[j][1];
	}
      
      fin = (sn+sn) > S0;
      i++;
    }
  /* transposition des resultats dans les variables de sorties */
  if(sn != S0)
    {
      moment0[0]     = sn;
      moment0[1]     = S0-sn;
      for(j=0;j<3;j++)
	{
	  moment1[0][j]     = En[j];
	  moment1[1][j]     = Moment1[j]-En[j];
	  moment2[0][j]     = En2[j];
	  moment2[1][j]     = Moment2[j]-En2[j];
	}
    }
  else
    {
      i--;
      moment0[0]     = sn-freq_axei;
      moment0[1]     = freq_axei;
      
      for(j=0;j<3;j++)
	{
	  moment1[0][j]     = En[j]-moments[j][0];
	  moment1[1][j]     = moments[j][0];
	  moment2[0][j]     = En2[j]-moments[j][1];
	  moment2[1][j]     = moments[j][1];
	}
    }
#ifdef DEBUG
  printf("en sortie de boucle i = %d\n",i);
  printf("gain= %d%\n",100*(b_sup1-i)/(b_sup1-b_inf1));
#endif
  return(i);
}

static int private_decoupe_median_dangereux(arbre_binaire self,
				     coul_freq tab_freq,
				     double* moment0,
				     moments_fils moment1,
				     moments_fils moment2)

{
  float S0   =  Moment0;
  int axe1   = (Axe_decoupe+1)%3;
  int axe2   = (axe1 + 1)%3;
  int b_inf1 = Inter[Axe_decoupe].borne_inf;
  int b_sup1 = Inter[Axe_decoupe].borne_sup;
#ifdef DEBUG
  int b_inf2 = Inter[axe1].borne_inf;
  int b_sup2 = Inter[axe1].borne_sup;
  int b_inf3 = Inter[axe2].borne_inf;
  int b_sup3 = Inter[axe2].borne_sup;
#endif
  int i,j; /* variables de boucle */
  double moments[3][2];
  bouleen fin = faux;
  double  En[3];
  double sn,terme,ui,freq_axei;
  int demi_intervalle = (REDUCTION>>1);
  
  /* initialisations */
  /************************************/
  
  i= b_inf1;
  sn =0.0;
  for(j=0;j<3;j++)
    En[j]=0.0;
  
#ifdef DEBUG
  printf("sur l'axe I%d : [%d,%d[\n",Axe_decoupe+1,b_inf1,b_sup1);
  printf("sur l'axe I%d : [%d,%d[\n",axe1+1,b_inf2,b_sup2);
  printf("sur l'axe I%d : [%d,%d[\n",axe2+1,b_inf3,b_sup3);
#endif
  /* Calculs */
  while((i<b_sup1-1)&&(!fin))
    {
      freq_axei= coul_freq_projection(tab_freq,Axe_decoupe,i,Inter,moments[axe1],moments[axe2]);
      
      
      ui = (i<<K_REDUCTION);
      if(i<0)
	ui -= demi_intervalle;
      else
	ui += demi_intervalle;
      
      sn += freq_axei;
      terme = ui*freq_axei;
      moments[Axe_decoupe][0] = terme;
      for(j=0;j<3;j++)
	En[j]   += moments[j][0];
      
      fin = (sn+sn) > S0;
      i++;
    }
  /* transposition des resultats dans les variables de sorties */
  if(sn != S0)
    {
      moment0[0]     = sn;
      moment0[1]     = S0-sn;
      for(j=0;j<3;j++)
	{
	  moment1[0][j]     = En[j];
	  moment1[1][j]     = Moment1[j]-En[j];
	  moment2[0][j]     = 0.0;
	  moment2[1][j]     = 0.0;
	  
	}
    }
  else
    {
      i--;
      moment0[0]     = sn-freq_axei;
      moment0[1]     = freq_axei;
      
      for(j=0;j<3;j++)
	{
	  moment1[0][j]     = En[j]-moments[j][0];
	  moment1[1][j]     = moments[j][0];
	  moment2[0][j]     = 0.0;
	  moment2[1][j]     = 0.0;
	}
    }
#ifdef DEBUG
  printf("en sortie de boucle i = %d\n",i);
  printf("gain= %d%\n",100*(b_sup1-i)/(b_sup1-b_inf1));
#endif
  return(i);
}

static double private_decoupe_choix(arbre_binaire self,coul_freq tab_freq)
{
  float S0 = Moment0;
  int axe1 = (Axe_decoupe+1)%3;
  int axe2 = (axe1 + 1)%3;
  int b_inf1 = Inter[Axe_decoupe].borne_inf;
  int b_sup1 = Inter[Axe_decoupe].borne_sup;

#ifdef DEBUG

  int b_inf2 = Inter[axe1].borne_inf;
  int b_sup2 = Inter[axe1].borne_sup;
  int b_inf3 = Inter[axe2].borne_inf;
  int b_sup3 = Inter[axe2].borne_sup;

#endif

  int i,j; /* variables de boucle */
  double  En[3],En2[3];
  double sn,terme,ui,freq_axei;
  double max = 0.0;
  double moments[3][2];
  double EQM;
  int demi_intervalle = (REDUCTION>>1);
  bouleen fin = faux;
  double X_moins_ya,det2;
  double a[3];
  double b2_par_delta2,delta2=0.0;
  /* initialisations */
  /************************************/
  /* sn : cardinal de Ct */
  /* En[], En2[] moments d'ordre 1 et 2*/
  /* a : moyenne du cluster */
  /* delta2 norme au carre du verteur tel que delta_i = longueur de */
  /* l'intervalle i */
  /************************************/
  sn=0.0;
  for(j=0;j<3;j++)
    {
      En[j]=0.0;
      En2[j] =0.0;
      a[j] = Moment1[j]/S0;
      delta2 += (Inter[j].borne_sup-Inter[j].borne_inf)*(Inter[j].borne_sup-Inter[j].borne_inf);
    }
  delta2 = delta2*REDUCTION*REDUCTION;
  b2_par_delta2 = S0*(S0/delta2);
  i= b_inf1;
#ifdef DEBUG
  printf("sur l'axe I%d : [%d,%d[\n",Axe_decoupe+1,b_inf1,b_sup1);
  printf("sur l'axe I%d : [%d,%d[\n",axe1+1,b_inf2,b_sup2);
  printf("sur l'axe I%d : [%d,%d[\n",axe2+1,b_inf3,b_sup3);
#endif
  /* Calculs */
  while((i<b_sup1)&&(!fin))
    {
      freq_axei= coul_freq_projection(tab_freq,Axe_decoupe,i,Inter,moments[axe1],moments[axe2]);
      
      
      ui = (i<<K_REDUCTION);
      if(i<0)
	ui -= demi_intervalle;
      else
	ui += demi_intervalle;
      
      sn += freq_axei;
      terme = ui*freq_axei;
      
      En[Axe_decoupe]   += terme;
      En[axe1] += moments[axe1][0];
      En[axe2] += moments[axe2][0];
      En2[Axe_decoupe] +=  ui*terme;
      En2[axe1] += moments[axe1][1];
      En2[axe2] += moments[axe2][1];
      EQM=0.0;
      if((sn != 0.0)&&(sn != S0))
	{
	  for(j=0;j<3;j++)
	    {
	      X_moins_ya = En[j]/sn-a[j];
	      EQM += X_moins_ya*X_moins_ya;
	    }
	  EQM = (EQM/(S0-sn))*sn;
	}
      
      if(max <  EQM)
	{
	  max = EQM;
	  det2 =b2_par_delta2*max;
	}
      fin = ((sn+sn) > S0)&&(sn*(S0-sn) <= det2);
      i++;
    }
  /* transposition des resultats dans les variables de sorties */
  
#ifdef DEBUG
  printf("en sortie de boucle i = %d\n",i);
  printf("gain= %d%\n",(int)(100*(b_sup1-i)/(b_sup1-b_inf1)));
#endif
  return(max);
}


   /*******************************************************************/
  /* Calcul des moments d'ordre 0, 1 et 2 d'un cluster. Utilise pour */
 /* l'initialisation                                                */
/*******************************************************************/
static void private_initialisation(arbre_binaire self, 
			    coul_freq tab_freq,
			    double* moment0, 
			    double* moment1, 
			    double* moment2)
{
  int i,j;
  int axeI1 = 0;
  int axeI2 = 1;
  int axeI3 = 2;
  int b_inf1 = Inter[axeI1].borne_inf;
  int b_sup1 = Inter[axeI1].borne_sup;
  double freq_axe_I,terme,sn;
  double En[3],En2[3];
  double mom[3][2];
  int ui,demi_intervalle = (REDUCTION>>1);
  
  
  sn=0.0;
  for(j=0;j<3;j++)
    {
      En[j]=0.0;
      En2[j]=0.0;
    }
  
/* Calcul de la somme; de la somme des carres; du cardinal de */
  /* l'ensemble */
  for(i=b_inf1;i<b_sup1;i++)
    {
      freq_axe_I=coul_freq_projection(tab_freq,axeI1,i,Inter,mom[axeI2],mom[axeI3]);
      ui = (i*REDUCTION);
      if(i<0)
	ui -= demi_intervalle;
      else 
	ui += demi_intervalle;

      sn += freq_axe_I;
      terme = ui*freq_axe_I;
      
      En[0] += terme;
		En2[0] = En2[0] + ui*terme;
      for(j=1;j<3;j++)
	{
	  En[j] += mom[j][0];
	  En2[j] += mom[j][1];
	  
	}
      
    }
  
  /* On transvase dans les variables de sortie */

  *moment0 = sn;
  for(j=0;j<3;j++)
    {
      moment1[j] = En[j];
      moment2[j] = En2[j];
    }
}
static void private_init_dangereux(arbre_binaire self, 
			    coul_freq tab_freq,
			    double* moment0, 
			    double* moment1, 
			    double* moment2)

{
  int i,j;
  int axeI1 = 0;
  int axeI2 = 1;
  int axeI3 = 2;
  int b_inf1 = Inter[axeI1].borne_inf;
  int b_sup1 = Inter[axeI1].borne_sup;
  double freq_axe_I,terme,sn;
  double En[3];
  double mom[3][2];
  int ui,demi_intervalle = (REDUCTION>>1);
  

  sn=0.0;
  for(j=0;j<3;j++)
    En[j]=0.0;
  
  /* Calcul de la somme; de la somme des carres; du cardinal de */
  /* l'ensemble */
  for(i=b_inf1;i<b_sup1;i++)
    {
      freq_axe_I=coul_freq_projection(tab_freq,axeI1,i,Inter,mom[axeI2],mom[axeI3]);
      ui = (i*REDUCTION);
      if(i<0)
	ui -= demi_intervalle;
      else 
	ui += demi_intervalle;
      
      sn += freq_axe_I;
      terme = ui*freq_axe_I;
      
      En[0] += terme;
      for(j=1;j<3;j++)
	En[j] += mom[j][0];
    }
  
  /* On transvase dans les variables de sortie */
  
  *moment0 = sn;
  for(j=0;j<3;j++)
    {
      moment1[j] = En[j];
      moment2[j] = 0.0;
    }
}

static arbre_binaire  private_creer_fils(arbre_binaire self)
{
  int i;
  arbre_binaire fils= (arbre_binaire)malloc(sizeof(struct arbre_binaire));
  
  fils->fils[0] = (arbre_binaire)NULL;
  fils->fils[1] = (arbre_binaire)NULL;
  
  for(i=0;i<3;i++)
    {
      fils->inter[i].borne_inf = Inter[i].borne_inf;
      fils->inter[i].borne_sup = Inter[i].borne_sup;
    }
  fils->feuille=vrai;

  fils->choix_axe = Choix_axe;
  fils->valeur = Valeur;
  fils->partition = Partition;

  return(fils);
}

/******************************************************/
/*choix de l'axe a couper */
/* choix_axe est egal a une de ces fonctions */
/******************************************************/

/* L'axe apportant la plus forte diminution de l'EQM */
static int private_choix_eqm(arbre_binaire self,
			     coul_freq tab_freq)
{
  double max_eqm,max =0.0;
  int axe_decoupe;
  int i;
  int longueur;

  for(i=0;i<3;i++)
    {
      longueur = Inter[i].borne_sup-Inter[i].borne_inf;
      Axe_decoupe = i;
      max_eqm = private_decoupe_choix(self,tab_freq);
      if((max_eqm > max)&&(longueur>1))
	{
	  axe_decoupe = i;
	  max = max_eqm;
	}
    }

  return(axe_decoupe);
}

/* L'axe apportant la plus forte diminution de la variance sur un axe */
static int private_choix_split_var(arbre_binaire self,
				   coul_freq tab_freq)
{
  double max_eqm,max =0.0;
  int axe_decoupe;
  int i,j,indice;
  int longueur;
  double diff_moy,X_moins_ya;
  double m0;
  moments_fils m1,m2;

  for(i=0;i<3;i++)
    {
      longueur = Inter[i].borne_sup-Inter[i].borne_inf;
      Axe_decoupe = i;
      indice = private_decoupe_projette(self,tab_freq,&m0,m1,m2);
      diff_moy=0.0;
      for(j=0;j<3;j++)
	{
	  X_moins_ya = m1[0][j]/m0 - Moment1[j]/Moment0;
	  diff_moy    += X_moins_ya*X_moins_ya;
	}
      if(Moment0 == m0)
	max_eqm = 0.0;
      else
	max_eqm = (m0/(Moment0-m0))*diff_moy;

      if((max_eqm > max)&&(longueur>1))
	{
	  axe_decoupe = i;
	  max = max_eqm;
	}
    }

  return(axe_decoupe);
}
/* L'axe de plus grande variance */
static int private_choix_var(arbre_binaire self,
			     coul_freq tab_freq)
{
  int i;
  double var_max= 0.0;
  int longueur;
  int axe_decoupe;
  
  for(i=0;i<3;i++)
    {
      longueur = Inter[i].borne_sup-Inter[i].borne_inf;
      if((Var[i] > var_max)&&(longueur>1))
	{
	  axe_decoupe = i;
	  var_max = Var[i];
	}
    }
  return(axe_decoupe);
}
/* L'axe le plus long */
static int private_choix_longueur(arbre_binaire self,
			   coul_freq tab_freq)
{
  int i;
  int longueur,longueur_max = 0;
  int axe_decoupe;
  
  for(i=0;i<3;i++)
    {
      longueur = Inter[i].borne_sup-Inter[i].borne_inf;
      if(longueur>longueur_max)
	{
	  axe_decoupe = i;
	  longueur_max = longueur;
	}
    }
  return(axe_decoupe);
}
/* L'axe tel que var(i)*longueur(i) soit max */
static int private_choix_var_long(arbre_binaire self,
			   coul_freq tab_freq)
{
  int i;
  int longueur;
  double mom2,mom2_max = 0.0;
  int axe_decoupe;
  
  for(i=0;i<3;i++)
    {
      longueur = Inter[i].borne_sup-Inter[i].borne_inf;
      mom2 = Var[i]*longueur;
      if((mom2>mom2_max)&&(longueur >1))
	{
	  axe_decoupe = i;
	  mom2_max = mom2;
	}
    }
  return(axe_decoupe);
}
/* L'axe tel que Var(i)*ecart_median(i) est max */
/* ecart_max(i) est la distance entre la moyenne sur l'axe et le point */
/* median sur cet axe. Avec pt_median(i) = ( Max(i)+Min(i) )/2 */
static int private_choix_var_median(arbre_binaire self,
                                    coul_freq tab_freq)
{
  int i;
  int longueur;
  int axe_decoupe;
  double mom2,mom2_max =0.0;
  int  min[3];
  int moy[3];
  int* median;
  

  for(i=0;i<DIM;i++)
    {
      moy[i] = Moy_axe[i]/REDUCTION;
      min[i] = Inter[i].borne_inf;
    } 
  median = coul_freq_give_median(tab_freq,min,moy);

  for(i=0;i<DIM;i++)
    {
      longueur = Inter[i].borne_sup-Inter[i].borne_inf;
      mom2 = Var[i]*(moy[i]-median[i]);
      if((mom2>mom2_max)&&(longueur >1))
	{
	  axe_decoupe = i;
	  mom2_max = mom2;
	}
    }
  return(axe_decoupe);
}


   /*******************************************/
  /* Calcul de l'homogneite du cluster       */
 /* valeur est egale a une de ces fonctions */
/*******************************************/

/* retourne l' EQM pondere ie :*/
/* La somme des distances au barycentre divise par le cardinal */
static double private_eqm_pondere(arbre_binaire self)
{
  int i;
  double erreur =0.0;
  
  for(i=0;i<3;i++)
    erreur += Var[i];

  return(erreur);
}
/* Retourne la somme des distances au barycentre */
static double private_eqm_non_pondere(arbre_binaire self)
{
  int i;
  double erreur =0.0;
  
  for(i=0;i<3;i++)
    erreur += Var[i];
  
  erreur = erreur*Moment0;

  return(erreur);
}
/* Retourne le cardinal */
static double private_cardinal(arbre_binaire self)
{
  int i;
  int longueur,longueur_max = 0;
  
  for(i=0;i<3;i++)
    {
      longueur = Inter[i].borne_sup-Inter[i].borne_inf;
      if(longueur>longueur_max)
	longueur_max = longueur;
    }

  if(longueur_max <= 1)
    return 0;
  
  return(Moment0);
}
/******************************************************/

static void private_creer_table(arbre_binaire self,
				coul_freq tab_freq)
{
  int i,j,indice_decoupe;
  arbre_binaire fils;
  double moment0[2];
  moments_fils moment1,moment2;
  double m0,m1,m2;
  
  if(Feuille)
    {
      
      indice_decoupe =Partition(self,tab_freq,moment0,moment1,moment2);
#ifdef DEBUG		
      printf("indice_decoupe suivant I%d: %d\n",Axe_decoupe+1,indice_decoupe);
#endif
      /* On corrige les intervalles */
      
      fils = Fils[0]= private_creer_fils(self);
      fils->inter[Axe_decoupe].borne_sup = indice_decoupe;
      
      fils = Fils[1]= private_creer_fils(self);
      fils->inter[Axe_decoupe].borne_inf = indice_decoupe;
      
      /* Boucle pour les deux fils (droit et gauche) */
      for(i=0;i<2;i++)
	{
	  fils = Fils[i];
	  m0 = fils->moment0 = moment0[i]; 
	  
	  /* boucle sur les 3 coordonnees */
	  for(j=0;j<3;j++)
	    {
	      m1 = fils->moment1[j]= moment1[i][j]; 
	      m2 = fils->moment2[j]= moment2[i][j]; 
	      fils->moy_axe[j]= m1/m0;
	      fils->var[j] = (m2-m1*m1/m0)/m0;
	      
#ifdef DEBUG
	      printf("var et moyenne sur l'axe I%d : var =%f moyenne = %d\n",(j+1),fils->var[j],fils->moy_axe[j]);
#endif						
	      if(fils->var[j] <-1.0)
		{
		  fprintf(stderr,"ERREUR DANS La quantification var negative\n");
		  exit(0);
		}
	    }
	  fils->axe_decoupe = Choix_axe(fils,tab_freq);
	  fils->ecart = Valeur(fils);
	}			    
      
      
      
      
#ifdef DEBUG
      printf("\n\n");
#endif
      /* On corrige le pere */
      Feuille=faux;
    }
  else
    {
      if( Fils[0]->ecart > Fils[1]->ecart )
	private_creer_table(Fils[0],tab_freq);
      else
	private_creer_table(Fils[1],tab_freq);
    }
  Ecart = MAX(Fils[0]->ecart,Fils[1]->ecart);
}

   /****************************************************************/
  /********** Remplie la table de couleurs  en prenant la couleur */
 /********** moyenne de chaque cluster dans l'arbre              */
/****************************************************************/
static void private_etiqueter(arbre_binaire self,int *table,int* numero)
{
  int i;
  int indice;
  
  if(Feuille)
    {
      indice = 3*(*numero);
      if(table != NULL)
	{
	  for(i=0;i<3;i++)
	    {
	      
	      *(table+indice)= Moy_axe[i];
	      indice++;
	    }
	}
      Numero = *numero;
      (*numero)++;
    }
  else
    {
      private_etiqueter(Fils[0],table,numero);
      private_etiqueter(Fils[1],table,numero);
    }
}












