/* border_value.c (Combinatorial Pyramids project)
 *  Copyright (C) 2003 Luc Brun
 *  email: luc.brun@univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "border_value.h"

static float *grad_x=NULL;
static float *grad_y=NULL;
static int    grad_dimx;
static int    grad_dimy;
static double alpha_deriche=0.9;

#define Max SELF(max)
#define Dart_value SELF(dart_value)
#define Dart_cardinal SELF(dart_cardinal)

struct border_value
{
  float max;
  tab_dart_value dart_value;
  tab_dart dart_cardinal;
};

extern border_value new_border_value(image I)
{
  float max;
  INSTANCIER(border_value);
  Dart_value = FAIRE_tab_dart_value();
  Dart_cardinal = FAIRE_tab_dart();
  init_deriche_deriv(alpha_deriche,I);
  grad_x=deriche_deriv_x_float(&grad_dimx, &grad_dimy, &Max);
  grad_y=deriche_deriv_y_float(&grad_dimx, &grad_dimy, &max);
  if (max>Max)
    Max=max;
  return self;
}

extern void destroy_border_value(border_value self)
{
  close_deriche_deriv();
  DEFAIRE_tab_dart_value(Dart_value);
  DEFAIRE_tab_dart(Dart_cardinal);
  if (grad_x != 0)
    free(grad_x);
  grad_x=0;
  if (grad_y != 0)
    free(grad_y);
  grad_y=0;
}
extern void border_value_put_alpha_deriche_value(double val)
{
  alpha_deriche=val;
}
extern float border_value_give_value(border_value self,int ind)
{
  return tab_dart_value_give_value(Dart_value, ind);
}
extern int border_value_give_cardinal(border_value self,int ind)
{
  return tab_dart_give_value(Dart_cardinal, ind);
}
extern float border_value_give_mean(border_value self,int ind)
{
  float val;
  int card;
  val=tab_dart_value_give_value(Dart_value, ind);
  card=tab_dart_give_value(Dart_cardinal, ind);
  return val/card;
}
extern float border_value_give_max(border_value self)
{
  return Max;
}
extern void  border_value_put_value(border_value self,int ind,float val)
{
  tab_dart_value_put_value(Dart_value, ind, val);
}
extern void  border_value_put_cardinal(border_value self,int ind,int card)
{
  tab_dart_put_value(Dart_cardinal, ind, card);
}

extern float init_edge(any_ptr data,pyramid P,dart d)
{
  int i;
  move_type dir;
  
  i=pyramid_give_vertex(P,d);
  dir=give_move(P,d);

  if (dir==right || dir==left)
    return grad_x[i];
  return grad_y[i];
}

extern void merge_border(any_ptr data,
			 pyramid P, 
                         int level,
                         dart d1,dart d2)
{
  border_value self=(border_value)data;
  int C, C_m;
  double M, M_m;
  M =tab_dart_value_give_value(Dart_value,d1);
  M+=tab_dart_value_give_value(Dart_value,d2);
  C =tab_dart_give_value(Dart_cardinal,d1);
  C+=tab_dart_give_value(Dart_cardinal,d2);

  M_m =tab_dart_value_give_value(Dart_value,-d1);
  M_m+=tab_dart_value_give_value(Dart_value,-d2);
  C_m =tab_dart_give_value(Dart_cardinal,-d1);
  C_m+=tab_dart_give_value(Dart_cardinal,-d2);

  tab_dart_value_put_value(Dart_value,d1,M);
  tab_dart_put_value(Dart_cardinal,d1,C);
  tab_dart_value_put_value(Dart_value,-d1,M_m);
  tab_dart_put_value(Dart_cardinal,-d1,C_m);
}
