/* deriche.c (Combinatorial Pyramids project)
 *  Copyright (C) 2003 Luc Brun
 *  email: luc.brun@univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "deriche.h"
#include <math.h>



static double  ema;
static double  Sa ;
static double  a  ;
static double  b1 ;
static double  b2 ;
static double  a0 ;
static double  a1 ;
static double  a2 ;
static double  a3 ;
static bouleen opened = faux;
static int     dim;
static int     dimx;
static int     dimy;
static image   img_orig;

static float *convert_image_priv();
static image  unconvert_image_priv(float*,float);

static float *deriv_x_priv(float*);
static float *deriv_y_priv(float*);

static float *make_norm_priv(float*,float*,float*);
static float* make_norm_one_priv(float*,float*);

static float * applique_filtre_h_x_priv(float*);
static float * applique_filtre_h_y_priv(float*);
static float * applique_filtre_g_x_priv(float*);
static float * applique_filtre_g_y_priv(float*);


int open_deriche_deriv(double alpha,char* I)
{

  img_orig = FAIRE_image();

  if(image_charger(img_orig,I) != 0)
    return -1;

  dim  = image_give_dim(img_orig);
  dimx = image_give_largeur(img_orig);
  dimy = image_give_hauteur(img_orig);


  ema = exp(-alpha); 
  Sa = -(1-ema)*(1-ema)/ema;
  a = Sa*ema;
  b1 = -2*ema;
  b2 = ema*ema;
  a0 = (1-ema)*(1-ema)/(1+2*alpha*ema-ema*ema);
  a1 = a0*(alpha-1)*ema;
  a2 = a1-a0*b1;
  a3 = -a0*b2;

  opened = vrai;
  
  return 0;
}
void init_deriche_deriv(double alpha,image I)
{
  img_orig=I;
  dim  = image_give_dim(img_orig);
  dimx = image_give_largeur(img_orig);
  dimy = image_give_hauteur(img_orig);

  ema = exp(-alpha); 
  Sa = -(1-ema)*(1-ema)/ema;
  a = Sa*ema;
  b1 = -2*ema;
  b2 = ema*ema;
  a0 = (1-ema)*(1-ema)/(1+2*alpha*ema-ema*ema);
  a1 = a0*(alpha-1)*ema;
  a2 = a1-a0*b1;
  a3 = -a0*b2;

  opened = vrai;
}
void close_deriche_deriv()
{
//  DEFAIRE_image(img_orig);
  opened = faux;
}
image deriche_deriv()
{
  float *I_float,*Gradx,*Grady,*norm_grad;
  float  max;
  image img_norm_grad;

  assert(opened);

  I_float   = convert_image_priv();

  Gradx     = deriv_x_priv(I_float);
  Grady     = deriv_y_priv(I_float);

  norm_grad = make_norm_priv(Gradx,Grady,&max);

  free(I_float);
  free(Gradx);
  free(Grady);
  
  img_norm_grad = unconvert_image_priv(norm_grad,max);
  free(norm_grad);
  
  return (img_norm_grad);
}
float *deriche_deriv_float(int *w, int *h, float *max)
{
  float *I_float,*Gradx,*Grady,*norm_grad;

  assert(opened);

  I_float   = convert_image_priv();

  Gradx     = deriv_x_priv(I_float);
  Grady     = deriv_y_priv(I_float);

  norm_grad = make_norm_priv(Gradx,Grady,max);

  free(I_float);
  free(Gradx);
  free(Grady);
  
  *w = dimx;
  *h = dimy;
  
  return (norm_grad);
}

image deriche_deriv_x()
{
  float *I_float,*Gradx,*norm_grad;
  float  max;
  image img_norm_grad;

  assert(opened);

  I_float   = convert_image_priv();

  Gradx     = deriv_x_priv(I_float);

  norm_grad = make_norm_one_priv(Gradx,&max);

  free(I_float);
  free(Gradx);
  
  img_norm_grad = unconvert_image_priv(norm_grad,max);
  free(norm_grad);
  
  return (img_norm_grad);
}
float *deriche_deriv_x_float(int *w, int *h, float *max)
{
  float *I_float,*Gradx,*norm_grad;

  assert(opened);

  I_float   = convert_image_priv();

  Gradx     = deriv_x_priv(I_float);

  norm_grad = make_norm_one_priv(Gradx,max);

  free(I_float);
  free(Gradx);
  
  return (norm_grad);
}

image deriche_deriv_y()
{
  float *I_float,*Grady,*norm_grad;
  float  max;
  image img_norm_grad;

  assert(opened);

  I_float   = convert_image_priv();

  Grady     = deriv_y_priv(I_float);

  norm_grad = make_norm_one_priv(Grady,&max);

  free(I_float);
  free(Grady);
  
  img_norm_grad = unconvert_image_priv(norm_grad,max);
  free(norm_grad);

  return (img_norm_grad);
}
float *deriche_deriv_y_float(int *w, int *h, float *max)
{
  float *I_float,*Grady,*norm_grad;

  assert(opened);

  I_float   = convert_image_priv();

  Grady     = deriv_y_priv(I_float);

  norm_grad = make_norm_one_priv(Grady,max);

  free(I_float);
  free(Grady);
  
  return (norm_grad);
}

static float* convert_image_priv()
{
  float *I,*I_cour;
  int* pix;
  int k;

  assert(opened);

  I = (float*) malloc(dim*sizeof(float)*dimx*dimy);

  I_cour = I;
  image_debut(img_orig);
  do
    {
      pix = image_lire_pixel(img_orig);
      for(k=0;k<dim;k++)
	{
	  *I_cour = pix[k];
	  I_cour++;
	}
    }
  while(image_pixel_suivant(img_orig));
  
  return (I);
}
static image unconvert_image_priv(float* norm_grad, float max)
{
  image img_returned;
  int pix;

  assert(opened);

  img_returned = FAIRE_image();
  image_initialize(img_returned,1,dimx,dimy);

  image_debut(img_returned);
  do
    {
      pix = (int)((255.0*(*norm_grad))/max);
      norm_grad++;
      image_ecrire_pixel(img_returned,&pix);
    }
  while(image_pixel_suivant(img_returned));
  
  return (img_returned);
}
static float* make_norm_one_priv(float* I,float *max)
{
  int i,j,k;
  float* norm_grad,*norm_grad_cour;
  float max_cour;
  
  norm_grad = (float*)malloc(dimx*dimy*sizeof(float));
  
  max_cour = 0.0;
  norm_grad_cour = norm_grad;
  for(i=0;i<dimy;i++)
    for(j=0;j<dimx;j++)
      {
	*norm_grad_cour = 0.0;
	for(k=0;k<dim;k++)
	  {
	    *norm_grad_cour += (*I)*(*I);
	    I++;
	  }
	*norm_grad_cour = sqrt((double)(*norm_grad_cour));
	if(*norm_grad_cour > max_cour)
	  max_cour = *norm_grad_cour;
	norm_grad_cour++;
      }
  *max = max_cour;
#ifdef DEBUG
  fprintf(stderr,"max_cour = %f\n",max_cour);
#endif
  return norm_grad;
}
static float* make_norm_priv(float* Dx,float* Dy,float* max_norm)
{
  double E,F,G;
  int i,j,k;	/* variables de boucles */
  float *normed_grad,*normed_grad_cour;
  float max;
  
  normed_grad = (float*)malloc(sizeof(float)*dimx*dimy);
  max = 0.0;
  normed_grad_cour = normed_grad;
  for(i=0;i<dimy;i++)
    {
      for(j=0;j<dimx;j++)
	{
	  E= G = F = 0.0;
	  for(k=0;k<dim;k++)
	    {
	      E += (*Dx)*(*Dx);
	      F += (*Dx)*(*Dy);
	      G += (*Dy)*(*Dy);
	      Dx++;
	      Dy++;
	    }
	  *normed_grad_cour = E+G+0.5*sqrt((E-G)*(E-G)+4*F*F);
	  *normed_grad_cour = sqrt(*normed_grad_cour);
	  if(max < *normed_grad_cour)
	    max = *normed_grad_cour; 
	  normed_grad_cour++;
	}
    }
  *max_norm = max;
#ifdef DEBUG
  fprintf(stderr,"max : %f",*max_norm);
#endif
  return normed_grad;
}
	    
static float* deriv_x_priv(float* I)
{
  float *S,*Grx;

  assert(opened);

  S = applique_filtre_g_x_priv(I);
  Grx = applique_filtre_h_y_priv(S);

  free(S);
  return Grx; 
}
static float* deriv_y_priv(float* I)
{
  float *S,*Gry;

  assert(opened);
  
  S = applique_filtre_h_x_priv(I);
  Gry = applique_filtre_g_y_priv(I);

  free(S);
  return Gry;
}
static float *applique_filtre_g_x_priv(float* I)
{
  float* I_cour;
  float* Yp,*Yp_cour,*Ym,*Ym_cour;
  float* Y1,*Y2;
  int next_line = dim*(dimx-1);
  int i,j,k;
  

  assert(opened);

  Yp = (float*) malloc(dim*sizeof(float)*dimx*dimy);
  Ym = (float*) malloc(dim*sizeof(float)*dimx*dimy);
  
  /* On initialise les 2 premieres colonnes */
  Yp_cour = Yp;
  I_cour = I;
  for(i=0;i<dimy;i++)
    {
      for(k=0;k<dim;k++)
	{
	  *Yp_cour = 0.0;
	  *(Yp_cour+dim) = *I_cour;
	  Yp_cour++;
	  I_cour++;
	}
      I_cour  += next_line;
      Yp_cour += next_line;
    }
  /* On s'occupe des autres colonnes */

  for(j=2;j<dimx;j++)
    {           
      Yp_cour = Yp + j*dim;			/* (0,j) */
      I_cour  = I  + (j-1)*dim;			/* (0,j-1) */
      Y1      = Yp + (j-1)*dim;			/* (0,j-1) */
      Y2      = Yp + (j-2)*dim;			/* (0,j-2) */
    
      for(i=0;i<dimy;i++)
	{
	  for(k=0;k<dim;k++)
	    {
	      *Yp_cour = *I_cour -b1*(*Y1)- b2*(*Y2);
	      Y2++;
	      Y1++;
	      Yp_cour++;
	      I_cour++;
	    }
	  Yp_cour += next_line;
	  Y1      += next_line;
	  Y2      += next_line;
	  I_cour  += next_line;
	}            
      
    }
  /* Calcul de Ym */
  /* On initialise les 2 dernieres colonnes */
  Ym_cour = Ym + dim*(dimx-1);
  I_cour = I   + dim*(dimx-1);
  for(i=0;i<dimy;i++)
    {
      for(k=0;k<dim;k++)
	{
	  *Ym_cour = 0.0;
	  *(Ym_cour-dim) = *I_cour;
	  Ym_cour++;
	  I_cour++;
	}
      I_cour  = I_cour + next_line;
      Ym_cour = Ym_cour+ next_line;
    }
  /* On s'occupe des autres colonnes */

  for(j=dimx-2;j>0;j--)
    {           
      Ym_cour = Ym + (j-1)*dim;			/* (0,dimx-j) */
      I_cour  = I  + j*dim;			/* (0,dimx-j+1) */
      Y1      = Ym + j*dim;			/* (0,dimx-j+1) */
      Y2      = Ym + (j+1)*dim;			/* (0,dimx-j+2) */
    
      for(i=0;i<dimy;i++)
	{
	  for(k=0;k<dim;k++)
	    {
	      *Ym_cour = *I_cour -b1*(*Y1)- b2*(*Y2);
	      Y2++;
	      Y1++;
	      Ym_cour++;
	      I_cour++;
	    }
	  Ym_cour += next_line;
	  Y1      += next_line;
	  Y2      += next_line;
	  I_cour  += next_line;
	}            
      
    }

  
  /* Calcul de S */
  Yp_cour = Yp;
  Ym_cour = Ym;
  
  for(j=0;j<dimy;j++)
    for(i=0;i<dimx;i++)
      for(k=0;k<dim;k++)
	{
	  *Yp_cour = a*((*Yp_cour)- (*Ym_cour));
	  Yp_cour++;
	  Ym_cour++;
	}
  free(Ym);
  return Yp;
}

static float *applique_filtre_g_y_priv(float* I)
{
  float* I_cour;
  float *Yp,*Yp_cour,*Ym,*Ym_cour;
  float *Y1,*Y2;
  int next_line = dim*dimx;
  int i,j,k;					/* variables de boucle */

  Yp = (float*) malloc(dim*sizeof(float)*dimx*dimy);
  Ym = (float*) malloc(dim*sizeof(float)*dimx*dimy);
  assert(Yp);
  assert(Ym);
  /* On initialise les 2 premieres lignes */
  Yp_cour = Yp;
  I_cour = I;
  for(i=0;i<dimx;i++)
    {
      for(k=0;k<dim;k++)
	{
	  *Yp_cour = 0.0;
	  *(Yp_cour + next_line) = *I_cour;
	  Yp_cour++;
	  I_cour++;
	}
    }
  /* On s'occupe des autres lignes */

  Yp_cour = Yp + 2*dim*dimx;             /* (i,0) */
  I_cour  = I  + dim*dimx;	       	/* (i-1,0) */
  Y1      = Yp + dim*dimx;	       /* (i-1,0) */
  Y2      = Yp;                       /* (i-2,0) */
  
  for(i=2;i<dimy;i++)
    {           
      for(j=0;j<dimx;j++)
	{
	  for(k=0;k<dim;k++)
	    {
	      *Yp_cour = *I_cour -b1*(*Y1)- b2*(*Y2);
	      Y2++;
	      Y1++;
	      Yp_cour++;
	      I_cour++;
	    }
	}            
      
    }
  /* Calcul de Ym */
  /* On initialise les 2 dernieres lignes */
  Ym_cour = Ym  + dim*(dimy-1)*dimx;		/* (dimy,0) */
  I_cour  = I   + dim*(dimy-2)*dimx;			/* (dimy-1,0) */
  for(i=0;i<dimx;i++)
    {
      for(k=0;k<dim;k++)
	{
	  *Ym_cour = 0.0;
	  *(Ym_cour - next_line) = *I_cour;
	  Ym_cour++;
	  I_cour++;
	}
    }
  /* On s'occupe des autres lignes */

  Ym_cour = Ym + dim*(dimy-2)*dimx-1;	/* (dimy-2,dimx) */
  I_cour  = I  + dim*(dimy-1)*dimx-1;	/* (dimy-1,dimx) */
  Y1      = Ym + dim*(dimy-1)*dimx-1;	/* (dimy-1,dimx) */
  Y2      = Ym + dim*dimy*dimx    -1;   /* (dimy,dimx) */
    
  for(i=dimy-2;i>0;i--)
    {           
      for(j=dimx-2;j>0;j--)
	{
	  for(k=0;k<dim;k++)
	    {
	      *Ym_cour = *I_cour -b1*(*Y1)- b2*(*Y2);
	      Y2--;
	      Y1--;
	      Ym_cour--;
	      I_cour--;
	    }
	}            
      
    }

  
  /* Calcul de S */
  Yp_cour = Yp;
  Ym_cour = Ym;
  
  for(j=0;j<dimy;j++)
    for(i=0;i<dimx;i++)
      for(k=0;k<dim;k++)
	{
	  *Yp_cour = a*(*Yp_cour-*Ym_cour);
	  Yp_cour++;
	  Ym_cour++;
	}

  free(Ym);
  return Yp;
}

static float *applique_filtre_h_x_priv(float* I)
{
  float *I_cour,*I1_cour;
  float *Yp,*Yp_cour,*Ym,*Ym_cour;
  float *Y1,*Y2;
  int next_line = dim*(dimx-1);
  int i,j,k;					/* variables de boucle */


  Yp = (float*) malloc(dim*sizeof(float)*dimx*dimy);
  Ym = (float*) malloc(dim*sizeof(float)*dimx*dimy);
  
  /* On initialise les 2 premieres colonnes */
  Yp_cour = Yp;
  I_cour = I;
  for(i=0;i<dimy;i++)
    {
      for(k=0;k<dim;k++)
	{
	  *Yp_cour = a0*(*I_cour);
	  *(Yp_cour+dim) = a0*(*(I_cour+dim))+ a1*(*I_cour)-b1*(*Yp_cour);
	  Yp_cour++;
	  I_cour++;
	}
      I_cour  += next_line;
      Yp_cour += next_line;
    }
  /* On s'occupe des autres colonnes */

  for(j=2;j<dimx;j++)
    {           
      Yp_cour = Yp + j*dim;			/* (0,j) */
      I_cour  = I  + j*dim;			/* (0,j) */
      I1_cour = I  + (j-1)*dim;			/* (0,j-1) */
      Y1      = Yp + (j-1)*dim;			/* (0,j-1) */
      Y2      = Yp + (j-2)*dim;			/* (0,j-2) */
    
      for(i=0;i<dimy;i++)
	{
	  for(k=0;k<dim;k++)
	    {
	      *Yp_cour = a0*(*I_cour) +a1*(*I1_cour)-b1*(*Y1)- b2*(*Y2);
	      Y2++;
	      Y1++;
	      Yp_cour++;
	      I_cour++;
	      I1_cour++;
	    }
	  Yp_cour += next_line;
	  Y1      += next_line;
	  Y2      += next_line;
	  I_cour  += next_line;
	  I1_cour += next_line;	  
	}            
      
    }
  /* Calcul de Ym */
  /* On initialise les 2 dernieres colonnes */
  Ym_cour = Ym + dim*(dimx-1);
  I_cour  = I  + dim*(dimx-1);
  for(i=0;i<dimy;i++)
    {
      for(k=0;k<dim;k++)
	{
	  *Ym_cour = 0.0;
	  *(Ym_cour-dim) = a2*(*I_cour);
	  Ym_cour++;
	  I_cour++;
	}
      I_cour  = I_cour + next_line;
      Ym_cour = Ym_cour+ next_line;
    }
  /* On s'occupe des autres colonnes */

  for(j=dimx-2;j>0;j--)
    {           
      Ym_cour = Ym + (j-1)*dim;			/* (0,dimx-j) */
      I_cour  = I  + j*dim;			/* (0,dimx-j+1) */
      I1_cour = I  + (j+1)*dim;			/* (0,dimx-j+2) */
      Y1      = Ym + j*dim;			/* (0,dimx-j+1) */
      Y2      = Ym + (j+1)*dim;			/* (0,dimx-j+2) */
    
      for(i=0;i<dimy;i++)
	{
	  for(k=0;k<dim;k++)
	    {
	      *Ym_cour = a2*(*I_cour)+a3*(*I1_cour) -b1*(*Y1)- b2*(*Y2);
	      Y2++;
	      Y1++;
	      Ym_cour++;
	      I_cour++;
	      I1_cour++;
	    }
	  Ym_cour += next_line;
	  Y1      += next_line;
	  Y2      += next_line;
	  I_cour  += next_line;
	  I1_cour += next_line;
	}            
      
    }

  
  /* Calcul de S */
  Yp_cour = Yp;
  Ym_cour = Ym;
  
  for(j=0;j<dimy;j++)
    for(i=0;i<dimx;i++)
      for(k=0;k<dim;k++)
	{
	  *Yp_cour = *Yp_cour+*Ym_cour;
	  Yp_cour++;
	  Ym_cour++;
	}
  
  free(Ym);
  return Yp;
}

static float *applique_filtre_h_y_priv(float* I)
{
  float* I_cour,*I1_cour;
  float *Yp,*Yp_cour,*Ym,*Ym_cour;
  float *Y1,*Y2;
  int next_line = dim*dimx;
  int i,j,k;					/* variables de */
						/* boucles */

  Yp = (float*) malloc(dim*sizeof(float)*dimx*dimy);
  Ym = (float*) malloc(dim*sizeof(float)*dimx*dimy);
  
  /* On initialise les 2 premieres lignes */
  Yp_cour = Yp;
  I_cour = I;
  for(i=0;i<dimx;i++)
    {
      for(k=0;k<dim;k++)
	{
	  *Yp_cour = a0*(*I_cour);
	  *(Yp_cour + next_line) = a0*(*(I_cour+next_line)) +a1*(*I_cour)-b1*(*Yp_cour);
	  Yp_cour++;
	  I_cour++;
	}
    }
  /* On s'occupe des autres lignes */

  Yp_cour = Yp + 2*dim*dimx;             /* (2,0) */
  I_cour  = I  + 2*dim*dimx;	       	/* (2,0) */
  I1_cour = I  + dim*dimx;		/* (1,0)*/
  Y1      = Yp + dim*dimx;	       /* (1,0) */
  Y2      = Yp;                       /* (0,0) */
  
  for(i=2;i<dimy;i++)
    {           
      for(j=0;j<dimx;j++)
	{
	  for(k=0;k < dim;k++)
	    {
	      *Yp_cour = a0*(*I_cour)+a1*(*I1_cour) -b1*(*Y1)- b2*(*Y2);
	      Y2++;
	      Y1++;
	      Yp_cour++;
	      I_cour++;
	      I1_cour++;
	    }
	}            
      
    }
  /* Calcul de Ym */
  /* On initialise les 2 dernieres lignes */
  Ym_cour = Ym + dim*(dimy-1)*dimx;		/* (dimy,0) */
  I_cour = I   + dim*(dimy-2)*dimx;		/* (dimy-1,0) */
  for(i=0;i<dimx;i++)
    {
      for(k=0;k<dim;k++)
	{
	  *Ym_cour = 0.0;
	  *(Ym_cour - next_line) = a2*(*I_cour);
	  Ym_cour++;
	  I_cour++;
	}
    }
  /* On s'occupe des autres lignes */

  Ym_cour = Ym + dim*(dimy-2)*dimx-1;	/* (dimy-2,dimx) */
  I_cour  = I  + dim*(dimy-1)*dimx-1;	/* (dimy-1,dimx) */
  I1_cour = I  + dim*dimy*dimx    -1;   /* (dimy,dimx) */
  Y1      = Ym + dim*(dimy-1)*dimx-1;	/* (dimy-1,dimx) */
  Y2      = Ym + dim*dimy*dimx    -1;   /* (dimy,dimx) */
    
  for(i=dimy-2;i>0;i--)
    {           
      for(j=dimx;j>0;j--)
	{
	  for(k=0;k<dim;k++)
	    {
	      *Ym_cour = a2*(*I_cour)+a3*(*I1_cour) -b1*(*Y1)- b2*(*Y2);
	      Y2--;
	      Y1--;
	      Ym_cour--;
	      I_cour--;
	      I1_cour--;
	    }
	}            
      
    }

  
  /* Calcul de S */
  Yp_cour = Yp;
  Ym_cour = Ym;
  
  for(j=0;j<dimy;j++)
    for(i=0;i<dimx;i++)
      for(k=0;k<dim;k++)
	{
	  *Yp_cour = *Yp_cour+*Ym_cour;
	  Yp_cour++;
	  Ym_cour++;
	}
  
  free(Ym);
  return Yp;
}









