/* para_segmentation.c (Combinatorial Pyramids project)
 *  Copyright (C) 2003 Luc Brun
 *  email: luc.brun@univ-reims.fr
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the version 2 of the GNU General Public
 *  License as published by the Free Software Foundation.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/
#include "para_segmentation.h"

static pyramid_value value_vertex=NULL;
static border_value value_edge=NULL;
static double seuil_lambda_min=220;
static double seuil_lambda_max=2500;
static double progrssion_lambda=0;
static double percent_grad_lambda=0.9;

#define MIN(a,b) ((a)<(b)?(a):(b))
#define MAX(a,b) ((a)<(b)?(b):(a))

static int give_cardinal(pyramid,dart);
static int give_gradient(pyramid,dart);

extern void init_para_segmentation(pyramid_value p_value, border_value b_value)
{
  value_vertex=p_value;
  value_edge=b_value;
}
extern void para_segmentation_put_seuil_lambda_min(double val)
{
  seuil_lambda_min=val;
}
extern void para_segmentation_put_seuil_lambda_max(double val)
{
  seuil_lambda_max=val;
}
extern void para_segmentation_put_progression_lambda(double val)
{
  progrssion_lambda=val;
}
extern void para_segmentation_put_gradient_percent_value(double val)
{
  percent_grad_lambda=val;
}

bouleen para_lambda(pyramid P,
		    int level,
		    vertex V1,
		    vertex V2,
		    dart d)
{
  int k, dim=pyramid_value_give_dim_img(value_vertex);
  double res1=0,res2,C1[3],C2[3], seuil_grad;
  
  seuil_grad=border_value_give_max(value_edge);
  seuil_grad*=percent_grad_lambda;
  
  pyramid_value_give_mean(value_vertex,V1,C1);
  pyramid_value_give_mean(value_vertex,V2,C2);

  for (k=0;k<dim;k++)
    res1+=(C1[k]-C2[k])*(C1[k]-C2[k]);

  res2 =give_gradient(P, d);
  res2+=give_gradient(P,-d);
  k =give_cardinal(P, d);
  res2=res2/(k+k);

if (progrssion_lambda<=0)
  return (res1<seuil_lambda_min && res2<seuil_grad);

  k=MAX(seuil_lambda_min*level*progrssion_lambda, seuil_lambda_min);
  k=MIN(k, seuil_lambda_max);
  return (res1<k && res2<seuil_grad);
}
double para_value(pyramid P,
		  vertex mu)
{
  return pyramid_value_give_se(value_vertex, mu);
}

double para_interest(pyramid P,
		     int level,
		     vertex V1,
		     vertex V2,
		     dart d)
{
  int k, p1, p2, front;
  double g, d1, d2;
  double g_max=border_value_give_max(value_edge);

  g =give_gradient(P, d);
  k =g_max*give_cardinal(P, d);
  d1=g/k;
  g =give_gradient(P,-d);
  d2=g/k;
  d1=MAX(d1,d2);

  p1=pyramid_value_give_perimetre(value_vertex, V1);
  p2=pyramid_value_give_perimetre(value_vertex, V2);
  front=give_cardinal(P, d);
  d2=(double)MIN(p1, p2)/(8.0*front);
  
  return d1*d2;
}

static int give_cardinal(pyramid P,dart b)
{
  dart phi_b,sig_b;
  dart phi2_b,phi_seg_b;
  int i;
  
  sig_b=pyramid_give_sigma(P,b);
  phi_b=pyramid_give_sigma(P,-b);
  phi2_b=pyramid_give_sigma(P,-phi_b);
  phi_seg_b=pyramid_give_sigma(P,-sig_b);
  
  if(phi2_b==b)
    {
      i=border_value_give_cardinal(value_edge, b);
      return i+border_value_give_cardinal(value_edge, phi_b);
    }
  if(phi_seg_b == -b)
    {
      i=border_value_give_cardinal(value_edge, b);
      return i+border_value_give_cardinal(value_edge, sig_b);
    }
  return border_value_give_cardinal(value_edge, b);
}

static int give_gradient(pyramid P,dart b)
{
  dart phi_b,sig_b;
  dart phi2_b,phi_seg_b;
  int i;
  
  sig_b=pyramid_give_sigma(P,b);
  phi_b=pyramid_give_sigma(P,-b);
  phi2_b=pyramid_give_sigma(P,-phi_b);
  phi_seg_b=pyramid_give_sigma(P,-sig_b);
  
  if(phi2_b==b)
    {
      i=border_value_give_value(value_edge, b);
      return i+border_value_give_value(value_edge, phi_b);
    }
  if(phi_seg_b == -b)
    {
      i=border_value_give_value(value_edge, b);
      return i+border_value_give_value(value_edge, sig_b);
    }
  return border_value_give_value(value_edge, b);
}
